#################################################################
#								#
#		Conformersplit v1.0 November 2002		#
#		 Institut of Technical Chemistry		#
#		      TU Clausthal-Germany			#
#		 Jens Reichel and Marco Drache			#
#								#
#################################################################

# Variables
# methode(atoms_all)		number of atoms in molecule
# methode(bonds_all)		number of bonds in molecule
# methode(elements_all)		list of elements in molecule
# methode(order_all)		list of atomic numbers in molecule
# methode(bond_order_all)	list of bond orders in molecule
# methode(conf_all)		number of conformers
# methode(conf_conv)		list of purged conformers
# atomnumber			current atom number
# conf(atomnumber)		current atomn umber in molecule
# conf(atomnumber,element)	element of current atom 
# conf(atomnumber,atomname)	atom label of current atom
# conf(atomnumber,atomicnumber)	atomic number of current atom
# conf(atomnumber,x_cart)	x-coordinate of current atom
# conf(atomnumber,y_cart)	y-coordinate of current atom
# conf(atomnumber,z_cart)	z-coordinate of current atom
# bondnum			current bond number
# bond(bondnum,atom1)		first atom of current bond 
# bond(bondnum,atom2)		second atom of current bond
# bond(bondnum,bondorder)	bond order of current bond
# i				current conformer
# conformercount		accepted conformer
# energy(i)			energy of current conformer
# elements : H Li B C N O F Na Mg Al Si P S Cl K Ca Br du
# du dummy atom

proc conf_extraction {filename log mol2} {

	if {$filename == ""} {
		
		$log insert end "\nCan't proceed ! No input file\n"
		return
	}
	
	array set methode {
	
		atoms_all 0
		bonds_all 0
		conf_all 0
		elements_all ""
		order_all ""
		bond_order_all ""
		conf_conv ""
	
	}
	set methode(filename) $filename
	set atomnumber 0	
	set i 0
	set stringall ""
	set abbrc 0
	
	set conffile [open $methode(filename) r]
	
	$log insert end "\nbegin conformer extraction of $methode(filename) ...\n\n"
	update idletasks

	while {1} {
		set pos1 [tell $conffile]
		set zeile [read $conffile]
		if {$pos1==[tell $conffile]} {
			incr abbrc
			if {$abbrc==10} {break}
		}
		append stringall $zeile
		seek $conffile 1 current
		incr i
	}
	close $conffile
	set spartan_es 0
	if {[regexp -all "GEOMETRY2" $stringall] > 0} {
		set spartan_es 1
	}
	
		
	if {([regexp -all "Conformer archive" $stringall] == 0)&&($spartan_es==0)} {
		tk_messageBox -title ERROR -message "file is not a conformer archive\n Unable to proceed!" -type ok
		return
	}
	
	set xyzfilename [format "%s.xyz" [file rootname $methode(filename)]]
	set xyzfile [open $xyzfilename w]
	close $xyzfile
	
	#evaluation of elements, their atomic numbers and labels
	set labelstart [string first "ATOMLABELS" $stringall]
	set labelend [string first "ENDATOMLABELS" $stringall]
	set labelstr [string range $stringall [expr $labelstart+10] [expr $labelend -1]]
	set labelstr [split $labelstr]
	set labellength [llength $labelstr]
	for {set n 0} {$n<$labellength} {incr n} {
		if {[regexp {([^\"])+} [lindex $labelstr $n] match]} {
			incr atomnumber
			set conf($atomnumber,atomname) $match
			regexp -all {([a-zA-Z])+} $conf($atomnumber,atomname) match
			set element $match
			set conf($atomnumber) $atomnumber
			set conf($atomnumber,element) $element
			
			switch  $element {
			
				H { set conf($atomnumber,atomicnumber) 1 }
				Li { set conf($atomnumber,atomicnumber) 3 }
				B { set conf($atomnumber,atomicnumber) 5 }
				C { set conf($atomnumber,atomicnumber) 6 }
				N { set conf($atomnumber,atomicnumber) 7 }
				O { set conf($atomnumber,atomicnumber) 8 }
				F { set conf($atomnumber,atomicnumber) 9 }
				Na { set conf($atomnumber,atomicnumber) 11 }
				Mg { set conf($atomnumber,atomicnumber) 12 }
				Al { set conf($atomnumber,atomicnumber) 13 }
				Si { set conf($atomnumber,atomicnumber) 14 }
				P { set conf($atomnumber,atomicnumber) 15 }
				S { set conf($atomnumber,atomicnumber) 16 }
				Cl { set conf($atomnumber,atomicnumber) 17 }
				K { set conf($atomnumber,atomicnumber) 19 }
				Ca { set conf($atomnumber,atomicnumber) 20 }
				Br { set conf($atomnumber,atomicnumber) 35 }
				I { set conf($atomnumber,atomicnumber) 53 }
				default {
					tk_messageBox -title Achtung -message "$element is not defined!\nAssignment of atomic number uncertain.\nPlease check it." -type ok
					set conf($atomnumber,atomicnumber) 0
				}
			}
			if {[regexp $element $methode(elements_all)] == 0} {
				lappend methode(elements_all) $element
			}
			if {[regexp $conf($atomnumber,atomicnumber) $methode(order_all)] == 0} {
				lappend methode(order_all) $conf($atomnumber,atomicnumber)
			}
		}
	}
	set methode(atoms_all) $atomnumber
	
	set bondstart [string first "HESSIAN" $stringall]
	set bondend [string first "ENDHESS" $stringall]
	set bondstr [string range $stringall [expr $bondstart+7] [expr $bondend -1]]
	set bondstr [split $bondstr]
	set bondlength [llength $bondstr]
	set bondind 0
	set bondincr 0
	set bondnum 0
	set methode(bond_order_all) ""

	for {set n 0} {$n<$bondlength} {incr n} {
		if {[lindex $bondstr $n] != ""} {
			incr bondind
			if {$bondind > $methode(atoms_all) } {
				incr bondincr
				switch -exact $bondincr {
					1 {
						incr bondnum
						set bond($bondnum) $bondnum
						set bond($bondnum,atom1) [lindex $bondstr $n]
					}
					2 {set bond($bondnum,atom2) [lindex $bondstr $n]}
					3 {
						set bondincr 0
						set bond($bondnum,bondorder) [lindex $bondstr $n]
						if {[regexp $bond($bondnum,bondorder) $methode(bond_order_all)] == 0} {
							lappend methode(bond_order_all) $bond($bondnum,bondorder)
						}
					}
				}
			}
		}
	}
	
	set methode(bonds_all) $bondnum
	set methode(conf_all) [regexp -all "GEOMETRY" $stringall]
	if {$spartan_es==1} {
		$log insert end "total number of Conformers : [expr $methode(conf_all)/2]\n"
	} else {	
		$log insert end "total number of Conformers : $methode(conf_all)\n"	
	}
	$log insert end "number of atoms in molecule : $methode(atoms_all)\n"
	$log insert end "elements in molecule : $methode(elements_all)\n"
	$log insert end "atomic numbers in molecule : $methode(order_all)\n"
	$log insert end "number of bonds in molecule : $methode(bonds_all)\n"
	$log insert end "bond orders in molecule : $methode(bond_order_all)\n\n"

	set sm [string first "GEOMETRY\n" $stringall]
	set em_all [string last "0.0" $stringall]
	set mstr [string range $stringall $sm [expr 2+$em_all]]

	#generation of characters (integer values) other than numbers . and - for extraction
	set str_abbruch ""
	for {set n 0} {$n <= 9} {incr n} {
		lappend	str_abbruch [format "%c" $n]
	}
	for {set n 11} {$n <= 31} {incr n} {
		lappend	str_abbruch [format "%c" $n]
	}
	for {set n 33} {$n <= 44} {incr n} {
		lappend	str_abbruch [format "%c" $n]
	}
	lappend	str_abbruch [format "%c" 47]
	
	for {set n 58} {$n <= 255} {incr n} {	
		lappend	str_abbruch [format "%c" $n]
	}
	set str_abbruch_length [llength $str_abbruch]
	
	set i 1
	set conformercount 1
	set methode(conf_conv) ""
	while {$i<=$methode(conf_all)} {
		set sm [string first "GEOMETRY" $mstr]
		set em [string first "ENERGY" $mstr]
		if {$spartan_es==1} {
			if {[string first "GEOMETRY2" $mstr]==$sm} {
				set mstr1 [string range $mstr $em [expr 2+$em_all]]
				set se [string first "ENERGY" $mstr1]
				set ee [string first "0.0" $mstr1]
				set mstr [string range $mstr1 $ee [expr 2+$em_all]]
				incr i
				continue
			}
		}				
				
		set st [string range $mstr [expr $sm+8] [expr $em-1]]
		if {[string first \x0 $st]!=-1} {
			lappend methode(conf_conv) $i
			
			#if hex 0 is in the string the cutting process begins
			# Cut from the first occurence of an character other than number . - 
			# to the last 0hex and reject it
			
			set sts [string first \x0 $st]
			for {set n 0} {$n < $str_abbruch_length} {incr n} {
				if {[string first [lindex $str_abbruch $n] $st]!=-1} {
					set stso [string first [lindex $str_abbruch $n] $st]
					if {$stso<$sts} {
						set sts $stso
					}
				}
			}
			set ste [string last \x0 $st]
			set st_1 [string range $st 0 [expr $sts-1]]
			set st_2 [string range $st [expr $ste+1] 10000]
			set stsum ""
			append stsum $st_1
			append stsum $st_2
			set st $stsum
		}
		set tmp 0
		set atomnumber 0
		set st_ind [split $st]
		set st_ind_length [llength $st_ind]
		for {set q 0} {$q < $st_ind_length} {incr q} {
			if {[lindex $st_ind $q] != ""} {
				incr tmp
				switch -exact $tmp {
					1 { incr atomnumber }
					2 {set conf($atomnumber,x_cart) [lindex $st_ind $q] }
					3 {set conf($atomnumber,y_cart) [lindex $st_ind $q] }
					4 {set conf($atomnumber,z_cart) [lindex $st_ind $q] ; set tmp 0 }
				}	
			}
		}		
		set mstr1 [string range $mstr $em [expr 2+$em_all]]
		set se [string first "ENERGY" $mstr1]
		set ee [string first "0.0" $mstr1]
		set energy($i) [string range $mstr1 [expr $se+15] [expr $ee-1]]
	
		# generation of some atom types of the Tripos force field (needed for the mol2 file)
		for {set j 1} {$j<=$methode(atoms_all) } {incr j} {
			set nsingle 0
			set ndouble 0
			set ntriple 0
			for {set k 1} {$k<=$methode(bonds_all)} {incr k} {
				if {($bond($k,atom1)==$j)||($bond($k,atom2)==$j)} {
					switch -exact $bond($k,bondorder) {
						1 {incr nsingle}
						2 {incr ndouble}
						3 {incr ntriple}
					}
				}			
			}
			
			# 
			switch -exact $conf($j,element) {
			
				H	{ set conf($j,fft) "H" }
				C	{	
						set conf($j,fft) "C.3"
						if {$ndouble>0} {
							set conf($j,fft) "C.2"
						}
						if {$ntriple>0} {
							set conf($j,fft) "C.1"
						}
				  	}
				N	{
						set conf($j,fft) "N.3"
						if {$nsingle == 4} { set conf($j,fft) "N.4" }
						if {$ndouble>0} {
							set conf($j,fft) "N.2"
						}
						if {$ntriple>0} {
							set conf($j,fft) "N.1"
						}
				  	}
				O	{	
						set conf($j,fft) "O.3"
						if {$ndouble>0} {
							set conf($j,fft) "O.2"
						}
				  	}
				S	{	
						set conf($j,fft) "S.3"
						if {$ndouble>0} {
							set conf($j,fft) "S.2"
						}
				  	}
				P	{ set conf($j,fft) "P.3" }
				F	{ set conf($j,fft) "F" }
				Cl	{ set conf($j,fft) "Cl" }
				Br	{ set conf($j,fft) "Br" }
				I	{ set conf($j,fft) "I" }
				Si	{ set conf($j,fft) "Si" }
				Na	{ set conf($j,fft) "Na" }
				K	{ set conf($j,fft) "I" }
				Ca	{ set conf($j,fft) "I" }
				Li	{ set conf($j,fft) "I" }
				Al	{ set conf($j,fft) "I" }
				default { 
					set conf($j,fft) "Du"
					if {$i == 1} {
						tk_messageBox -title Achtung -message "atom $j do not map to Sybyl MOL2!\nAssignment of type \"Dummy\"" -type ok
					}	
				}
			}
		}
		set mstr [string range $mstr1 $ee [expr 2+$em_all]]
		
		# write mol2-file of first conformer
		if {$i == 1} {
			write_mol_file conf $energy($i) bond $methode(atoms_all) $methode(bonds_all) $conformercount $methode(filename)
			$log insert end "\nconformer 1 written as mol2-file\n"
		}
		# write mol2-files of conformers if desired
		if {$mol2 == 1 && $i > 1} {
			write_mol_file conf $energy($i) bond $methode(atoms_all) $methode(bonds_all) $conformercount $methode(filename)
		}
		#write one xyz-file containing all conformers
		write_xyz_file $methode(atoms_all) conf $energy($i) $conformercount $methode(filename)
		incr conformercount
		incr i
	}
	
	$log insert end "Converted conformers:  [expr $conformercount-1]\n"
	$log insert end "\nall [expr $conformercount-1] conformers written in [format "%s.xyz" [file rootname $methode(filename)]]"
	if {$mol2 == 1} {
		$log insert end "\nAlso written all conformers in mol2-files.\n"
	}
}


#quit program
proc quit {} {
	update idletasks
	set answer [tk_messageBox -title "Exit program" -message "Do you want to leave the program?" -type yesno]
       	switch -- $answer {
		yes {exit}
		no {return}
	}
}

# specify file for process
proc file_load {filename dat_path log} {

	set auswahl 0
	set types {
		{{input files}	{.spartan}	}
		{{all files}		*.*	}
	}
	set filename [tk_getOpenFile -title "Choose file for process" -defaultextension dat -filetypes $types -initialdir $dat_path]
	if {$filename == ""} {
		$log insert end "\ncurrent file: no input file\n"
	} else {
	$log insert end "\ncurrent file: $filename\n"
	}
	return $filename
}

# write MOL2 file
proc write_mol_file {conf_t energy bond_t atoms_all bonds_all i filename} {

	upvar $conf_t conf $bond_t bond
	set outputfile [format "%s_$i.mol2" [file rootname $filename]]
	set ch2 [open $outputfile w]
	
	puts $ch2 "#\n#\tFile Created by Conformersplit v1.0\n#\n"
	puts $ch2 "@<TRIPOS>MOLECULE\n$outputfile\n$atoms_all $bonds_all\nSMALL\nNO_CHARGES\n"
	puts $ch2 "@<TRIPOS>ATOM"
	
	for {set n 1} {$n <= $atoms_all} {incr n} {
		puts $ch2 [format "   %d\t%s\t%s\t%s\t%s\t%s" $conf($n) $conf($n,atomname) $conf($n,x_cart) $conf($n,y_cart) $conf($n,z_cart) $conf($n,fft)]
	}
	puts $ch2 "\n\n@<TRIPOS>BOND"
	for {set n 1} {$n <= $bonds_all} {incr n} {
		puts $ch2 [format "   %d   %d    %d    %s" $bond($n) $bond($n,atom1) $bond($n,atom2) $bond($n,bondorder)]
	}
	puts $ch2 "@<TRIPOS>COMMENT"
	puts $ch2 [format "Energy: %s" $energy]

	close $ch2	
}

# write file of all conformers in xyz file format
proc write_xyz_file {atoms_all conf_t energy number filename} {
	upvar $conf_t conf
	
	set xyzfilename [format "%s.xyz" [file rootname $filename]]
	set xyzfile [open $xyzfilename a]
		
	puts $xyzfile $atoms_all
	puts $xyzfile [format "molecule_%u\t%s" $number $energy]
	
	for {set n 1} {$n <= $atoms_all} {incr n} {
		puts $xyzfile [format " %s\t%s\t%s\t%s" $conf($n,element) $conf($n,x_cart) $conf($n,y_cart) $conf($n,z_cart)]
	}
	close $xyzfile	
}


################################################################
#	         generation of window			       #
################################################################

set dat_path [pwd]
set filename ""


wm iconname . Conformersplit
wm title . {   --------------------     Conformersplit v1.0 2002    --------------------}

frame .one -borderwidth 10
button .one.get -text "Open file" -underline 0 -command {set filename [file_load $filename $dat_path $log]}
button .one.start -text "process file" -underline 0 -command {conf_extraction $filename $log $mol2}
button .one.end -text "Quit" -underline 0 -command {quit}
button .one.about -text "About..." -command {tk_messageBox -title "About Conformersplit" -message "Conformersplit v1.0 November 2002\nCopyright (C) Jens Reichel and Marco Drache\nInstitut of Technical Chemistry\nTU Clausthal\
\nThis program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version.
This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA." -type ok}

checkbutton .one.mols -text "Generate MOL2-files" -variable mol2 -borderwidth 0

pack .one.get .one.start .one.end .one.mols .one.about -side left -fill x -expand true
pack .one -fill x -expand true

frame .t
set log [text .t.log -width 80 -height 10 -borderwidth 2 -relief raised -setgrid true \
	-yscrollcommand {.t.scroll set}]
scrollbar .t.scroll -command {.t.log yview}
pack .t.scroll -side right -fill y
pack .t.log -side left -fill both -expand true
pack .t -side top -fill both -expand true

bind . <Control-Key-o> {set filename [file_load $filename $dat_path $log]}
bind . <Alt-Key-p> {conf_extraction $filename $log $mol2}
bind . <Alt-Key-F4> {quit; break}
$log insert end "\nConformersplit version 1.0, Copyright (C) 2002  Jens Reichel, Marco Drache
Conformersplit comes with ABSOLUTELY NO WARRANTY.\nThis is free software, and you are welcome to redistribute it under certain conditions.\n"

wm withdraw . ; wm deiconify .

